#include "drawable-node.h"
#include <string.h>
#include "wsint.h"

#include <X11/Xlib.h>
#include <X11/extensions/Xfixes.h>

#if G_BYTE_ORDER == G_LITTLE_ENDIAN
#   define PIXEL_TYPE GL_UNSIGNED_BYTE
#else
#   define PIXEL_TYPE GL_UNSIGNED_INT_8_8_8_8_REV
#endif

static void drawable_node_update_pixmap (DrawableNode *node);

static guint
power_of_two_greater_than (guint in)
{
    guint result = 1;
    
    while (result < in)
	result *= 2;
    
    return result;
}

static WsShmPixmap *
get_shm_pixmap (WsDrawable *drawable,
		int	    width,
		int	    height,
		WsFormat    format)
{
    static WsShmPixmap *cache;
    static int cached_width;
    static int cached_height;
    static WsFormat cached_format;
    
    if (!cache || cached_width < width ||
	cached_height < height || cached_format != format)
    {
	if (cache)
	    ws_shm_pixmap_unref (cache);
	
	cache = ws_shm_pixmap_new (drawable, width, height, format);

	/* FIXME: we should do this properly at some point */
	g_assert (cache);
	
	cached_width = width;
	cached_height = height;
	cached_format = format;
    }
    
    return cache;
}

static gboolean
has_child_prefixed (WsWindow *window,
		    char *child_title)
{
    GList *list;
    GList *children = ws_window_query_subwindows (window);
    int child_len = strlen (child_title);
    
    for (list = children; list ; list = list->next)
    {
	char *title = ws_window_query_title (list->data);
	
	if (title && strncmp (child_title, title, child_len) == 0)
	    return TRUE;
    }
    
    return FALSE;
}

static void
fill_junk (guint32 *mem, gsize size)
{
    int i;
    
    for (i = 0; i < size/4; ++i)
	mem[i] = g_random_int ();
}

static int counter;
WsWindow *debug_window;

static guchar *
create_buffer (WsShmPixmap *pixmap, WsFormat format,
	       guint width, guint height)
{
    guchar *buffer;
    const guchar *src;
    int i, j;
    guint bytes_per_row;
    guint bytes_per_pixel;
    guint pot_width, pot_height;

    bytes_per_row = ws_shm_pixmap_get_bytes_per_row (pixmap);
    src = ws_shm_pixmap_get_data (pixmap);

    /* FIXME */
    bytes_per_pixel = 4; 

    pot_width = power_of_two_greater_than (width);
    pot_height = power_of_two_greater_than (height);
    
    buffer = g_malloc0 (4 * pot_height * pot_width);
#if 0
    fill_junk ((guint32 *)buffer, 4 * pot_height * pot_width);
#endif
    
    for (j = 0; j < height; ++j)
    {
	const guchar *src_row = src + (j % height) * bytes_per_row;
	guint32 *dest_row = (guint32 *)(buffer + j * 4 * pot_width);
	
	for (i = 0; i < width; ++i)
	{
	    const guchar *pixel = src_row + (i % width) * bytes_per_pixel;

	    dest_row[i] = *(guint32 *)pixel;

	    if (format == WS_FORMAT_RGB_32 ||
		format == WS_FORMAT_RGB_24)
	    {
		dest_row[i] |= 0xff000000;
	    }

#if 0
	    if ((dest_row[i] & 0xFF000000) == 0xFF000000)
		g_print ("alpha\n");
	    
	    dest_row[i] = (0xff << 24) |
		
	    dest_row[i] =  0x77ffffff & dest_row[i];
	    
	    if (i == 1 && j == 1)
	    {
		if (debug_window && has_child_prefixed (debug_window, "X-Chat"))
		    g_print ("pixel: %lx\n", dest_row[i]);
	    }
#endif
	}
    }
    
    return buffer;
}

static void
print_region (WsRegion *region, const char *name)
{
    XRectangle *rects;
    int i, n_rects;
    
    rects = XFixesFetchRegion (region->ws->xdisplay, region->xregion, &n_rects);
    g_print ("region \"%s\":\n", name);
    for (i = 0; i < n_rects; ++i)
	g_print ("  %d %d %d %d\n", rects[i].x, rects[i].y, rects[i].width, rects[i].height);
    XFree (rects);
}

static void
create_texture (DrawableNode *node,
		WsRectangle  *geometry)
{
    WsFormat format;
    guchar *buffer;
    WsShmPixmap *pixmap;
    
    /* FIXME assuming it's a window here */
    ws_begin_error_trap (node->drawable->ws);
    
    format = ws_window_query_format ((WsWindow *)node->drawable);

    pixmap = get_shm_pixmap ((WsDrawable *)node->pixmap,
			     geometry->width,
			     geometry->height,
			     format);

    ws_drawable_copy_area ((WsDrawable *)node->pixmap, 0, 0,
			   geometry->width,
			   geometry->height,
			   (WsDrawable *)pixmap, 0, 0, NULL);

    /* We need a sync here so that we will know that the data has been
     * copied. FIXME: could most likely be done more efficiently.
     *
     * We may even need a getimage type thing to make sure the stuff
     * has been flushed to the card.
     */
    ws_sync (ws_drawable_get_ws ((WsDrawable *)node->pixmap));
    
    glGenTextures (1, &node->texture);
    node->we_have_a_texture = 0xc001;
    g_assert (node->texture != 0);
    
    glBindTexture (GL_TEXTURE_2D, node->texture);
    glPixelStorei (GL_UNPACK_ALIGNMENT, 4);

    debug_window = (WsWindow *)node->drawable;
    
    buffer = create_buffer (pixmap, format, geometry->width, geometry->height);

    ws_end_error_trap (node->pixmap->drawable.ws);
    
    /* These are global I think so should be set in a more global place */
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_CLAMP);
    
    glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA,
		  power_of_two_greater_than (geometry->width),
		  power_of_two_greater_than (geometry->height), 0 /* border */,
		  GL_BGRA, PIXEL_TYPE, buffer);
    
    g_free (buffer);
}

static void
drawable_repair (DrawableNode *node)
{
    WsDrawable *drawable = node->drawable;
    WsRectangle geometry;
    gboolean get_all;
    
    /* FIXME: maybe at some point we should get
     * smart and do incremental update.
     */

    if (node->need_new_pixmap)
    {
	if (node->pixmap)
	    ws_pixmap_free (node->pixmap);
	
	ws_begin_error_trap (drawable->ws);
	node->pixmap = ws_window_name_pixmap ((WsWindow *)drawable);
	ws_end_error_trap (drawable->ws);

	node->need_new_pixmap = FALSE;
    }
    else
    {
	if (!node->damaged_region)
	    return;
    }


    if (node->texture)
    {
	glDeleteTextures (1, &node->texture);
	node->texture = 0;
	node->we_have_a_texture = 0xBAD;
    }
    else
    {
	g_assert (!(node->we_have_a_texture == 0xC001));
	g_assert (node->we_have_a_texture == 0xBAD);
    }

    ws_begin_error_trap (drawable->ws);
    
    if (!ws_drawable_query_geometry ((WsDrawable *)node->pixmap, &geometry))
	node->pixmap = NULL;

    ws_end_error_trap (drawable->ws);

    if (node->pixmap)
	create_texture (node, &geometry);

    if (node->damaged_region)
	ws_region_unref (node->damaged_region);
    
    node->damaged_region = NULL;
    
    node_queue_repaint ((Node *)node);
}

static void
on_damage (WsDrawable *drawable,
	   gpointer    data)
{
    DrawableNode *node = data;
    WsRegion *region;

    ws_begin_error_trap (drawable->ws);
    
    region = ws_drawable_undamage (drawable);

#if 0
    if (has_child_prefixed ((WsWindow *)drawable, "X-Chat"))
    {
	g_print ("damaged region is\n");

	print_region (region, "xchat");
    }
#endif
    
    ws_end_error_trap (drawable->ws);

    if (node->damaged_region)
    {
	ws_region_union (node->damaged_region, region);
	ws_region_unref (region);
    }
    else
    {
	node->damaged_region = region;
    }
}

static void
deform (DrawableNode *node,
	double time,
	double inx,
	double iny,
	double *outx,
	double *outy)
{
    if (node->deformation_func)
    {
	DrawableDeformationFunc deformation_func = node->deformation_func;
	gpointer data = node->deformation_data;
	
	deformation_func (time, inx, iny, outx, outy, data);
    }
    else
    {
	*outx = inx;
	*outy = iny;
    }
}

static void
drawable_node_render (Node *node)
{
    DrawableNode *drawable_node = (DrawableNode *)node;
    WsRectangle rect;
    WsScreen *screen;
    double scr_w, scr_h;
    double tw, th;
    double vx, vy;
    double vw, vh;
    gboolean compositable = TRUE;

    ws_begin_error_trap (drawable_node->drawable->ws);
    
    if (!ws_window_query_viewable ((WsWindow *)drawable_node->drawable) &&
	drawable_node->need_new_pixmap)
    {
	compositable = FALSE;
    }

    if (ws_window_query_input_only ((WsWindow *)drawable_node->drawable))
	compositable = FALSE;

    ws_end_error_trap (drawable_node->drawable->ws);

    if (!compositable || !drawable_node->viewable)
	return;

    ws_screen_graphics_sync (
	ws_drawable_query_screen (drawable_node->drawable));
    
    drawable_repair (drawable_node);
    
    ws_begin_error_trap (drawable_node->drawable->ws);
    
    ws_drawable_query_geometry (drawable_node->drawable, &rect);
    screen = ws_drawable_query_screen (drawable_node->drawable);

    ws_end_error_trap (drawable_node->drawable->ws);

    if (!screen)
	return;
    
    scr_w = ws_screen_get_width (screen);
    scr_h = ws_screen_get_height (screen);

    tw = rect.width / (double)power_of_two_greater_than (rect.width);
    th = rect.height / (double)power_of_two_greater_than (rect.height);

    vw = rect.width / scr_w;
    vh = rect.height / scr_h;

    vx = (rect.x / scr_w);
    vy = rect.y / scr_h;

    if (drawable_node->texture)
    {
#define N_GRID_POINTS 64
	int i, j;
	gdouble time = g_timer_elapsed (drawable_node->timer, NULL);

	glBindTexture (GL_TEXTURE_2D, drawable_node->texture);
	
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor4f (1.0, 1.0, 1.0, 1.0);
	
	for (i = 0; i < N_GRID_POINTS; ++i)
	{
	    for (j = 0; j < N_GRID_POINTS; ++j)
	    {
		double tx0, ty0, tx1, ty1;
		double wx0, wy0, wx1, wy1;

		/* The transformed points */
		double p0x, p0y;
		double p1x, p1y;
		double p2x, p2y;
		double p3x, p3y;
		    

		tx0 = (tw / N_GRID_POINTS) * i;
		tx1 = (tw / N_GRID_POINTS) * (i + 1);
		ty0 = (th / N_GRID_POINTS) * j;
		ty1 = (th / N_GRID_POINTS) * (j + 1);

		wx0 = vx + (vw / N_GRID_POINTS) * i;
		wx1 = vx + (vw / N_GRID_POINTS) * (i + 1);
		wy0 = 1.0 - (vy + (vh / N_GRID_POINTS) * j);
		wy1 = 1.0 - (vy + (vh / N_GRID_POINTS) * (j + 1));

		deform (drawable_node, time, wx0, wy0, &p0x, &p0y);
		deform (drawable_node, time, wx1, wy0, &p1x, &p1y);
		deform (drawable_node, time, wx1, wy1, &p2x, &p2y);
		deform (drawable_node, time, wx0, wy1, &p3x, &p3y);
		
		glBegin (GL_QUADS);
		glTexCoord2f (tx0, ty0); glVertex3f (p0x, p0y, 0.0);
		glTexCoord2f (tx1, ty0); glVertex3f (p1x, p1y, 0.0);
		glTexCoord2f (tx1, ty1); glVertex3f (p2x, p2y, 0.0);
		glTexCoord2f (tx0, ty1); glVertex3f (p3x, p3y, 0.0);
		glEnd ();
	    }
	}
#if 0
	
	glBegin (GL_QUADS);
	glTexCoord2f (0.0, 0.0); glVertex3f   (vx, 1.0 - vy, 0.0);
	glTexCoord2f (tw, 0.0);  glVertex3f   (vx + vw, 1.0 - vy, 0.0);
	glTexCoord2f (tw, th);   glVertex3f   (vx + vw, 1.0 - vy - vh, 0.0);
	glTexCoord2f (0.0, th);  glVertex3f   (vx, 1.0 - vy - vh, 0.0);
	glEnd ();

	glDisable (GL_TEXTURE_2D);
#endif
    }
}

static void
drawable_node_compute_extents (Node    *node,
			       Extents *extents)
{
    extents->x = 0.0;
    extents->y = 0.0;
    extents->z = 0.0;
    extents->width = 1.0;
    extents->height = 1.0;
    extents->depth = 1.0;
}

#include "wsint.h"

static void
on_configure (WsWindow *drawable,
	      int x,
	      int y,
	      int width,
	      int height,
	      int bw,
	      WsWindow *above,
	      gboolean override_redirect,
	      gpointer data)
{
    DrawableNode *node = data;
#if 0
    g_print ("configure\n");
    
    g_print ("%d %d %d %d\n", geometry.x, geometry.y,
	     geometry.width, geometry.height);
#endif
    
    node->need_new_pixmap = TRUE;
}

DrawableNode *
drawable_node_new (WsDrawable *drawable)
{
    DrawableNode *node;
    WsRectangle rectangle;

    g_return_val_if_fail (drawable != NULL, NULL);

    node = g_new0 (DrawableNode, 1);

#if 0
    g_print ("drawable node new %p\n", node);
#endif

    node->node.render = drawable_node_render;
    node->node.compute_extents = drawable_node_compute_extents;

    node->drawable = drawable;
    node->texture = 0;
    node->we_have_a_texture = 0xBAD;

    ws_init_composite (ws_drawable_get_ws (drawable));
    
    /* FIXME: we are assuming it here again */
    ws_begin_error_trap (drawable->ws);

    node->need_new_pixmap = TRUE;
    node->pixmap = NULL;

    node->timer = g_timer_new ();
    
    ws_end_error_trap (drawable->ws);
    
    node->damaged_region = NULL;
    
    ws_init_damage (ws_drawable_get_ws (drawable));
    
    ws_drawable_set_damage_callback (drawable, on_damage, node);

    ws_window_set_configure_callback ((WsWindow *)drawable, on_configure, node);

    ws_drawable_query_geometry (drawable, &rectangle);

    node->viewable = TRUE;
    
    return node;
}

void
drawable_node_set_viewable (DrawableNode *node,
			    gboolean	  viewable)
{
    node->viewable = viewable;

    node_queue_repaint ((Node *)node);
}

void
drawable_node_unref (DrawableNode *node)
{
    ws_drawable_free (node->drawable);

    if (node->texture)
    {
#if 0
	g_print ("destroying out of unref\n");
#endif
	glDeleteTextures (1, &node->texture);
	node->texture = 0;
	node->we_have_a_texture = 0xBAD;
    }

    if (node->damaged_region)
    {
	ws_region_unref (node->damaged_region);
	node->damaged_region = NULL;
    }

#if 0
    node_unref ((Node *)node);
#endif
}

void
drawable_node_set_deformation_func (DrawableNode *		node,
				    DrawableDeformationFunc	func,
				    gpointer			data)
{
    node->deformation_func = func;
    node->deformation_data = data;
}
						 
