#include <glib.h>

#include <X11/X.h>
#include <X11/Xutil.h>
#include <X11/extensions/Xcomposite.h>
#include <X11/extensions/Xdamage.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "scene.h"

#if G_BYTE_ORDER == G_LITTLE_ENDIAN
#   define PIXEL_TYPE GL_UNSIGNED_BYTE
#else
#   define PIXEL_TYPE GL_UNSIGNED_INT_8_8_8_8_REV
#endif

struct SGCoord
{
    gboolean top;
    
    union
    {
	struct {
	    /* FIXME: maybe factor some of this out into a separate
	     * compositing manager
	     */
	    WsWindow  *root;
	    WsWindow  *gl_window;
	} top;
	
	struct {
	    SGCoord *parent;
	    
	    /* Matrix transform */    // transforms parent coordinates into
				      // our coordinates
	} sub;
    } u;
    
    GList *windows;
};

struct SGVertex
{
    double x;
    double y;
};

static WsShmPixmap *
get_shm_pixmap (WsDrawable *drawable,
		int	    width,
		int	    height,
		WsFormat    format)
{
    static WsShmPixmap *cache;
    static int cached_width;
    static int cached_height;
    static WsFormat cached_format;
    
    if (!cache || cached_width < width ||
	cached_height < height || cached_format != format)
    {
#if 0
	const WsColor blue = { 0xFFFF, 0x0000, 0x0000, 0xFFFF } ;
#endif
	
	if (cache)
	    ws_shm_pixmap_unref (cache);
	
	cache = ws_shm_pixmap_new (drawable, width, height, format);
	
	g_assert (cache);
	
	cached_width = width;
	cached_height = height;
	cached_format = format;
	
#if 0
	ws_drawable_fill_area (pixmap, 0, 0, -1, -1, &blue);
#endif
    }
    
    return cache;
}

#include "wsint.h"

static void
convert_pixel (WsFormat format,
	       const guchar *src,
	       guint *a,
	       guint *r,
	       guint *g,
	       guint *b)
{
    switch (format)
    {
    case WS_FORMAT_RGB_16:
    {
	guint16 t = *(guint16 *)src;
	guint tr, tg, tb;
	
	tr = t & 0xf800;
	*r = (tr >> 8) + (tr >> 13);
	
	tg = t & 0x07e0;
	*g = (tg >> 3) + (tg >> 9);
	
	tb = t & 0x001f;
	*b = (tb << 3) + (tb >> 2);
	
	*a = 0xff;
    }
    break;
    
    case WS_FORMAT_RGB_24:
    {
	*r = src[0];
	*g = src[1];
	*b = src[2];
	*a = 0xff;
    }
    break;
    
    case WS_FORMAT_RGB_32:
    {
	guint32 t = *(guint32 *)src;
	*r = (t >> 16) & 0xff;
	*g = (t >> 8) & 0xff;
	*b = t & 0xff;
	*a = 0xff;
    }
    break;
    
    case WS_FORMAT_ARGB_32:
    {
	g_print ("BIRNAN\n");
    }
    break;
    
    default:
    {
	g_assert_not_reached();
	break;
    }
    
    }
}

static void
fill_junk (guint32 *mem, gsize size)
{
    int i;
    
    for (i = 0; i < size/4; ++i)
	mem[i] = 0xff000000;
#if 0
	mem[i] = (i % 4 == 3)? 0x10 : 0xaa ; /* ((i % 4) * (i % 100)) */
#endif
}

static guint
power_of_two_greater_than (guint in)
{
    guint result = 1;
    
    while (result < in)
	result *= 2;
    
    return result;
}

static guchar *
create_buffer (WsShmPixmap *pixmap, guint width, guint height)
{
    guchar *buffer;
    WsFormat format;
    const guchar *src;
    int i, j;
    guint bytes_per_row;
    guint bytes_per_pixel;
    guint pot_width, pot_height;
    
    format = ws_shm_pixmap_get_format (pixmap);
    bytes_per_row = ws_shm_pixmap_get_bytes_per_row (pixmap);
    src = ws_shm_pixmap_get_data (pixmap);
    
    switch (format)
    {
    case WS_FORMAT_RGB_16:
	bytes_per_pixel = 2;
	break;
	
    case WS_FORMAT_RGB_24:
	bytes_per_pixel = 3;
	break;
	
    case WS_FORMAT_RGB_32:
    case WS_FORMAT_ARGB_32:
	bytes_per_pixel = 4;
	break;
	
    default:
	g_assert_not_reached();
	break;
    }
    
    pot_width = power_of_two_greater_than (width);
    pot_height = power_of_two_greater_than (height);
    
    buffer = g_malloc0 (4 * pot_height * pot_width);
    fill_junk ((guint32 *)buffer, 4 * pot_height * pot_width);
    
    for (j = 0; j < height; ++j)
    {
	const guchar *src_row = src + (j % height) * bytes_per_row;
	guint32 *dest_row = (guint32 *)(buffer + j * 4 * pot_width);
	
	for (i = 0; i < width; ++i)
	{
	    const guchar *pixel = src_row + (i % width) * bytes_per_pixel;
	    
	    dest_row[i] = 0xff << 24 | *(guint32 *)pixel;
	}
    }
    
    return buffer;
}

static gboolean
dump_error (const char *context)
{
    /* glGetError() is a server roundtrip */
    GLenum err;
    
    err = glGetError ();
    if (err != GL_NO_ERROR)
    {
	const GLubyte *message = gluErrorString (err);
	g_printerr ("GL Error: %s [at %s]\n", message, context);
	return TRUE;
    }
    
    return FALSE;
}

static GLuint
make_junk_texture ()
{
    static GLuint texture;
    static gboolean did_it;
    
    if (!did_it)
    {
	guint width = power_of_two_greater_than (500);
	guint height = power_of_two_greater_than (500);
	guchar *buffer = g_malloc (width * height * 4);
	
	fill_junk ((guint32 *)buffer, width * height * 4);
	
	glGenTextures (1, &texture);
	
	glBindTexture (GL_TEXTURE_2D, texture);
#if 0
	glPixelStorei (GL_UNPACK_ALIGNMENT, 4);
#endif
	
	glTexImage2D (GL_TEXTURE_2D, 0, 4,
		      width, height, 0 /* border */,
		      GL_BGRA, PIXEL_TYPE, buffer);
	
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	
	g_free (buffer);
	
	did_it = 1;
    }
    
    return texture;
}

static GHashTable *textures;

static GLuint
make_texture (WsWindow *window)
{
    GLuint texture;
    
    if (!textures)
	textures = g_hash_table_new (g_direct_hash, g_direct_equal);
    
    texture = GPOINTER_TO_UINT (g_hash_table_lookup (textures, window));
    
    if (!texture)
    {
	WsShmPixmap *pixmap;
	WsRectangle rect;
	WsFormat format;
	guchar *buffer;
	
	glGenTextures (1, &texture);
	if (texture == 0)
	    texture = 1;		/* we use 0 for none */
	
	ws_drawable_query_geometry ((WsDrawable *)window, &rect);
	format = ws_window_query_format (window);
	
	pixmap = get_shm_pixmap ((WsDrawable *)window, rect.width, rect.height, format);
	
	ws_drawable_copy_area ((WsDrawable *)window, 0, 0, rect.width, rect.height,
			       (WsDrawable *)pixmap, 0, 0, None);
	
	ws_sync (ws_drawable_get_ws ((WsDrawable *)window));
	
	buffer = create_buffer (pixmap, rect.width, rect.height);

	glBindTexture (GL_TEXTURE_2D, texture);

#if 0
	glPixelStorei (GL_UNPACK_ROW_LENGTH, rect.width);
#endif
	glPixelStorei (GL_UNPACK_ALIGNMENT, 4);
	
	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA,
		      power_of_two_greater_than (rect.width),
		      power_of_two_greater_than (rect.height), 0 /* border */,
		      GL_BGRA, PIXEL_TYPE, buffer);
	
	dump_error ("teximage\n");
	
	/* These are global I think so should be set in a more global place */
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	
	g_free (buffer);
	
	g_hash_table_insert (textures, window, GUINT_TO_POINTER (texture));
    }
    
    return texture;
}

typedef struct
{
    float tu, tv;
    float x, y, z;
} Vertex;

static Vertex cube_vertices [] =
{
    { 0.0f,0.0f,  0.0f, 0.0f, 0.0f },
    { 1.0f,0.0f,  1.0f, 0.0f, 0.0f },
    { 1.0f,1.0f,  1.0f, 1.0f, 0.0f },
    { 0.0f,1.0f,  0.0f, 1.0f, 0.0f },
   
    { 1.0f,0.0f,  0.0f, 0.0f,-1.0f },
    { 1.0f,1.0f,  0.0f, 1.0f,-1.0f },
    { 0.0f,1.0f,  1.0f, 1.0f,-1.0f },
    { 0.0f,0.0f,  1.0f, 0.0f,-1.0f },
   
    { 0.0f,1.0f,  0.0f, 1.0f,-1.0f },
    { 0.0f,0.0f,  0.0f, 1.0f, 0.0f },
    { 1.0f,0.0f,  1.0f, 1.0f, 0.0f },
    { 1.0f,1.0f,  1.0f, 1.0f,-1.0f },
   
    { 1.0f,1.0f,  0.0f, 0.0f,-1.0f },
    { 0.0f,1.0f,  1.0f, 0.0f,-1.0f },
    { 0.0f,0.0f,  1.0f, 0.0f, 0.0f },
    { 1.0f,0.0f,  0.0f, 0.0f, 0.0f },
   
    { 1.0f,0.0f,  1.0f, 0.0f,-1.0f },
    { 1.0f,1.0f,  1.0f, 1.0f,-1.0f },
    { 0.0f,1.0f,  1.0f, 1.0f, 0.0f },
    { 0.0f,0.0f,  1.0f, 0.0f, 0.0f },
   
    { 0.0f,0.0f,  0.0f, 0.0f,-1.0f },
    { 1.0f,0.0f,  0.0f, 0.0f, 0.0f },
    { 1.0f,1.0f,  0.0f, 1.0f, 0.0f },
    { 0.0f,1.0f,  0.0f, 1.0f,-1.0f }
};

GTimer *timer;
static int angle = 0;

static void
draw_window (WsWindow *window, gboolean as_cube)
{
    WsRectangle rect;
    double x, y, w, h;
    GLuint texture;
    double sw, sh;
    gint width;
    gint height;
    
    texture = make_texture (window); //  (window);
    glBindTexture (GL_TEXTURE_2D, texture);

    ws_drawable_query_geometry ((WsDrawable *)window, &rect);

    if (!ws_window_query_viewable (window) ||
	ws_window_query_input_only (window))
    {
	return;
    }

    width = rect.width;
    height = rect.height;
    rect.width = power_of_two_greater_than (rect.width);
    rect.height = power_of_two_greater_than (rect.height);

    sw = (double)ws_screen_get_width (ws_drawable_query_screen ((WsDrawable *)window));
    sh = (double)ws_screen_get_height (ws_drawable_query_screen ((WsDrawable *)window));
    
    x = (rect.x / sw) * 2 - 1;
    y = -((rect.y / sh) * 2 - 1);
    w = (rect.width / sw) * 2;
    h = (rect.height / sh) * 2;

    if (as_cube)
    {
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity();
	double w, h;
	
	glEnable (GL_DEPTH_TEST);
	
	gluPerspective( 45.0f, 1.0, 0.1f, 10.0f );

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	glTranslatef (x, y, -2.4);

	w = (width / sw) * 2;
	h = (height / sh) * 2;
    
	glScalef (w, -h, 1.0);

	glRotatef (angle, 0.5, 1.0, 0.5);

	double wt = width / (double)rect.width;
	double ht = height / (double)rect.height;
	
	/* Draw the cube */
	glBegin (GL_QUADS);
	glTexCoord2f (0.0f, 0.0f); glVertex3f (0.0f, 0.0f, 0.0f);
	glTexCoord2f (  wt, 0.0f); glVertex3f (1.0f, 0.0f, 0.0f);
	glTexCoord2f (  wt,   ht); glVertex3f (1.0f, 1.0f, 0.0f);
	glTexCoord2f (0.0f,   ht); glVertex3f (0.0f, 1.0f, 0.0f);
   
	glTexCoord2f (  wt, 0.0f); glVertex3f (0.0f, 0.0f, -1.0f);
	glTexCoord2f (  wt,   ht); glVertex3f (0.0f, 1.0f, -1.0f);
	glTexCoord2f (0.0f,   ht); glVertex3f (1.0f, 1.0f, -1.0f);
	glTexCoord2f (0.0f, 0.0f); glVertex3f (1.0f, 0.0f, -1.0f);
   
	glTexCoord2f (0.0f,   ht); glVertex3f (0.0f, 1.0f, -1.0f);
	glTexCoord2f (0.0f, 0.0f); glVertex3f (0.0f, 1.0f,  0.0f);
	glTexCoord2f (  wt, 0.0f); glVertex3f (1.0f, 1.0f,  0.0f);
	glTexCoord2f (  wt,   ht); glVertex3f (1.0f, 1.0f, -1.0f);
   
	glTexCoord2f (  wt,   ht); glVertex3f (0.0f, 0.0f, -1.0f);
	glTexCoord2f (0.0f,   ht); glVertex3f (1.0f, 0.0f, -1.0f);
	glTexCoord2f (0.0f, 0.0f); glVertex3f (1.0f, 0.0f,  0.0f);
	glTexCoord2f (  wt, 0.0f); glVertex3f (0.0f, 0.0f,  0.0f);
   
	glTexCoord2f (  wt, 0.0f); glVertex3f (1.0f, 0.0f, -1.0f);
	glTexCoord2f (  wt,   ht); glVertex3f (1.0f, 1.0f, -1.0f);
	glTexCoord2f (0.0f,   ht); glVertex3f (1.0f, 1.0f,  0.0f);
	glTexCoord2f (0.0f, 0.0f); glVertex3f (1.0f, 0.0f,  0.0f);
   
	glTexCoord2f (0.0f, 0.0f); glVertex3f (0.0f, 0.0f, -1.0f);
	glTexCoord2f (  wt, 0.0f); glVertex3f (0.0f, 0.0f,  0.0f);
	glTexCoord2f (  wt,   ht); glVertex3f (0.0f, 1.0f,  0.0f);
	glTexCoord2f (0.0f,   ht); glVertex3f (0.0f, 1.0f, -1.0f);
	glEnd ();
    }
    else
    {
	glDisable (GL_DEPTH_TEST);
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	
	glTranslatef (x, y, 0.0);
	glScalef (w, -h, 1.0);
	glBegin (GL_QUADS);
	
	glTexCoord2f (0.0, 0.0);
	glVertex3f   (0.0, 0.0, 0.0);
	
	glTexCoord2f (1.0,   0.0);
	glVertex3f   (1.0, 0.0, 0.0);
	
	glTexCoord2f (1.0,   1.0);
	glVertex3f   (1.0, 1.0, 0.0);
	
	glTexCoord2f (0.0, 1.0);
	glVertex3f   (0.0, 1.0, 0.0);
	
	glEnd ();
    }
}

static void
repaint (SGCoord *coord)
{
    glMatrixMode (GL_MODELVIEW);
    
    if (coord->top)
    {
	GList *list;
	
	ws_window_raise (coord->u.top.gl_window);
	
	glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable (GL_BLEND);
		
	glClearColor (0.0, 0.0, 0.0, 1.0);
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	glEnable (GL_TEXTURE_2D);
	
	for (list = coord->windows; list; list = list->next)
	{
	    WsWindow *window = list->data;
	    char *title = ws_window_query_title (window);

	    if (title)
	    {
		draw_window (window, FALSE);
	    }
	}

	glClear (GL_DEPTH_BUFFER_BIT);
	
	for (list = coord->windows; list; list = list->next)
	{
	    WsWindow *window = list->data;
	    char *title = ws_window_query_title (window);

	    make_texture (window); //  (window);
	    
	    if (!title)
		draw_window (window, TRUE);
	}
	
	ws_window_gl_swap_buffers (coord->u.top.gl_window);
	
	glFlush ();
	
	ws_flush (ws_drawable_get_ws ((WsDrawable *)coord->u.top.gl_window));
    }
    else
    {
	/* FIXME */
    }
}

static void
setup_for_compositing (SGCoord *coord)
{
    Ws *ws = ws_drawable_get_ws ((WsDrawable *)coord->u.top.root);
    
    if (!ws_init_composite (ws))
    {
	/* FIXME: report an error */
	g_assert (FALSE);
    }
    
    ws_window_redirect_subwindows (coord->u.top.root);
    
    /* Create GL window */
    coord->u.top.gl_window = ws_window_new_gl (coord->u.top.root);
    
    ws_window_set_override_redirect (coord->u.top.gl_window, TRUE);
    
    ws_window_map (coord->u.top.gl_window);
    
    
    ws_window_unredirect (coord->u.top.gl_window);
    
    ws_sync (ws_drawable_get_ws ((WsDrawable *)coord->u.top.gl_window));
    
    /* Get all windows */
    
}

SGVertex *
sg_vertex_new (SGCoord *coord, double x, double y)
{
    SGVertex *vertex = g_new0 (SGVertex, 1);
    
    vertex->x = x;
    vertex->y = y;
    
    return vertex;
}

SGCoord *
sg_coord_new_root (WsWindow *root)
{
    SGCoord *coord;
    
    g_return_val_if_fail (root != None, NULL);
    
    coord = g_new0 (SGCoord, 1);
    coord->top = TRUE;
    
    coord->u.top.root = root;
    
    coord->windows = NULL;
    
    setup_for_compositing (coord);
    
    repaint (coord);
    
    /* select for mapnotify */
    /* list windows */
    
    /* for each window:
       emit "new window"
    */
    return coord;
}

void
sg_window_attach (Window window,
		  gint x, gint y,
		  SGVertex *vertex)
{
    /* programming error if point (on window) is the same
     * as an existing one 
     */
    
    /* maintain set of lines and points */
    
    /* for all lines,
     * if it intersects the x of the new point,
     * if it is closer in the y direction then the previous
     */
    
    
    /* compute the grid positions and add new vertices as necessary:
     *
     * when drawing:
     * 
     *    first sort the point in reading order
     *  
     */
    
    /* horizontal:
     *
     * if the point closest
     */
}

static int idle_id;

static gboolean
do_repaint (gpointer data)
{
    SGCoord *coord = data;
    
#if 0
    g_print ("doing paint\n");
#endif
    
    repaint (coord);
    
#if 0
    g_print ("done painting\n");
#endif
    
idle_id = 0;
    
    return FALSE;
}

static void
queue_repaint (SGCoord *coord)
{
#if 0
    g_print ("queueing paint\n");
#endif
    
    if (!idle_id)
	idle_id = g_idle_add (do_repaint, coord);
}

static void
on_damage (WsDrawable *drawable,
	   gpointer data)
{
    SGCoord *coord = data;
    WsRegion *region = ws_drawable_undamage (drawable);
    WsRectangle rect;
    WsShmPixmap *shmpix;
    WsFormat format;
    GLuint texture;
    
    if (!textures)
	textures = g_hash_table_new (g_direct_hash, g_direct_equal);
    
    texture = GPOINTER_TO_UINT (g_hash_table_lookup (textures, drawable));
    if (texture)
    {
	glDeleteTextures (1, &texture);
	g_hash_table_remove (textures, drawable);
    }
    
    ws_drawable_query_geometry (drawable, &rect);
    format = ws_window_query_format ((WsWindow *)drawable);
    
    shmpix = get_shm_pixmap (drawable, rect.width, rect.height, format);
    
    g_assert (shmpix);
    
    /* acquire a texture */
    
    ws_region_unref (region);
    
    queue_repaint (coord);
}

void
sg_coord_add_window (SGCoord *coord,
		     SGVertex *v1,
		     SGVertex *v2,
		     SGVertex *v3,
		     SGVertex *v4,
		     WsWindow *window)
{
    WsDrawable *drawable = (WsDrawable *)window;
    
    if (!ws_init_damage (ws_drawable_get_ws (drawable)))
    {
	g_assert (FALSE);
	return;
    }
    
    if (window == coord->u.top.gl_window)
    {
	g_warning ("ignoring GL window\n");
	return;
    }
    
    ws_drawable_set_damage_callback (drawable, on_damage, coord);
    
    coord->windows = g_list_prepend (coord->windows, window);
    
    queue_repaint (coord);
    
    /* FIXME */
}

static gboolean
update_angle (gpointer data)
{
    SGCoord *coord = data;

    angle = g_timer_elapsed (timer, NULL) * 360;

    queue_repaint (coord);
}

void
scene_graph_start_animation(SGCoord *coord)
{
    static int id;

    if (!id)
    {
	timer = g_timer_new ();
    
	id = g_idle_add (update_angle, coord);
    }
}

void
scene_graph_new (const char *dpy)
{
    /* Become a compositing manager for dpy */
    
    /* Redirect root window */
    
    /* Create a GL window 
       
    - override redirect
    
    - raise to top
    
    */
    
    /* unredirect GL window */
    
    
    /* install event handlers 
       
    - on window create (for the root window)
    - on configure
    
    */
}
